//---------------------------------------------------------------------------

#include "stdafx.h"

#ifdef _MSC_VER
#pragma hdrstop
#endif

#include "kernelincludes.h"

#include "orthogonalize.h"
//---------------------------------------------------------------------------

Ortho::Ortho( PLA *_pla )
{
	pla = _pla;
    ok_terms = new PLA(pla);
	if ( ok_terms == NULL) throw RuntimeException( "Ortho: Not enough memory" );
	ok_terms->setUnique();
}

void Ortho::Orthogonalize(void)
/* Randomly select one of two intersecting terms, randomly splits one
*/
{
    string s, s1;

    Term newterm;
    list<Term>::iterator ti1, ti2, ti3, ti4;
    vector<int> candidates;
	vector<int>::iterator ci;
    unsigned int pos;
    unsigned int i, j;
    list<Term> *data;

	data = pla->getData();
			// first preprocessing
	RemoveAbsorbed();
	pla->MergeTerms();
	MoveOKTerms2();

    ti1 = data->begin();
    while ( ti1 != data->end()) {
       ti2 = ti1;
	   ti2++;
	   do {
            if ( ti2 == data->end() ) break;
			j = (*ti1).Relation2(*ti2);
			if ( j == 0 || (j & 8) == 0 ) ti2++;				// no IM or OM intersection
			else if ( j == 63 || j == 27 )
				ti2 = pla->DeleteTerm(ti2);		// equal or contained
			else if ( (j & 7) == 7 ) {				// equal inputs, different outputs. Merge them
                (*ti1).Merge(*ti2);
                ti2 = pla->DeleteTerm(ti2);
			} else if ( (j & 7) != 0 && (j & 8) != 0 ) break;   // non-empty IM and OM intersection
			else throw RuntimeException("Orthogonalize: Something's wrong");
       } while ( ti2 != data->end() );
       if ( ti2 != data->end()) {       // they do intersect
            for ( i = 0; i < pla->inputs(); i++ )
                if ( (*ti1).isLit_dc(i) && !(*ti2).isLit_dc(i) ) candidates.push_back(i + 1);
                else if ( (*ti2).isLit_dc(i) && !(*ti1).isLit_dc(i) ) candidates.push_back(-(int)(i + 1));
            if ( candidates.size() == 0 ) throw RuntimeException("Orthogonalize: Something's wrong");
            else {
				i = rand() % candidates.size();
				if ( candidates[i] > 0 ) ti3 = ti1; else ti3 = ti2;
                pos = abs(candidates[i]) - 1;
                candidates.clear();
                newterm = *ti3;         // make a copy
                (*ti3).setLit_0(pos);   // set 0 in the original term
                ti4 = data->begin();
				do {
					if ( ti4 != ti3 ) { 
                        if ( (*ti4).contains(*ti3) ) {
                            pla->DeleteTerm(ti3);
							ti1 = data->begin();
                            break;
                        } else if ( (*ti3).contains(*ti4) ) {
                            ti4 = pla->DeleteTerm(ti4);
							ti1 = data->begin();
                        } else ti4++;                    
					} else ti4++;
				} while ( ti4 != data->end() );
                newterm.setLit_0(pos);    // set 1 in the new term
                for ( ti4 = data->begin(); ti4 != data->end(); ti4++ ) {
                        if ( (*ti4).contains(newterm) ) break;	// already included somewhere
                }
                if ( ti4 == data->end() ) pla->AppendTerm(newterm);
            }
         //   pla->SavePLA("tmp.pla");
       } else ti1++;
    }
	MoveOKTerms2();
	if ( pla->getTerms() != 0 ) throw RuntimeException( "Ortho::Orthogonalize:: Something's wrong - resulting PLA is not orthogonal" );
}

void Ortho::MoveOKTerms(void)
// Moves terms intersecting with no other terms to ok_terms. Matrix must be computed
{
	unsigned int i, j;
	list<Term>::iterator ti;

	for ( i = 0; i < pla->getTerms(); i++ ) {
		for ( j = 0; j < pla->getTerms(); j++ )
			if ( matrix[i][j] != -1 ) break;
		if ( j == pla->getTerms() ) {
			ti = pla->getTermByIndex(i);
			ok_terms->AppendTerm(*ti);
			(*ti).setUnwanted();
		}
	}
	pla->Purge();
}

void Ortho::MoveOKTerms2(void)
// Moves terms intersecting with no other terms to ok_terms. Does not rely on matrix.
{
	list<Term>::iterator ti1, ti2;
	bool flag;

	for ( ti1 = pla->getData()->begin(); ti1 != pla->getData()->end(); ti1++ ) {
		flag = true;
		for ( ti2 = pla->getData()->begin(); ti2 != pla->getData()->end(); ti2++ ) 
			if ( ti1 != ti2 )
				if ( (*ti1).intersects(*ti2)) {
					flag = false;
					break;
				}
		if ( flag ) {
			ok_terms->AppendTerm(*ti1);
			(*ti1).setUnwanted();
		}
	}
	pla->Purge();
}

void Ortho::GenerateMatrix(void)
{
	vector<int> matrixrow;
    unsigned int i, j, k;
    list<Term>::iterator ti1, ti2;

	matrix.clear();
	matrixrow.clear();
	matrixrow.resize(pla->getTerms(), 0);	
	for ( i = 0; i < pla->getTerms(); i++ ) 
		matrix.push_back(matrixrow);
	for ( i = 0; i < pla->getTerms(); i++ ) {
		matrix[i][i] = -1;
		ti1 = pla->getTermByIndex(i);
		for ( j = i+1; j < pla->getTerms(); j++ ) {
			ti2 = pla->getTermByIndex(j);
			k = (*ti1).intersectSize(*ti2);
			matrix[i][j] = k;
			matrix[j][i] = k;
		}
	}
}

void Ortho::RecomputeMatrix(void)
{
    unsigned int i, j, k;
    list<Term>::iterator ti1, ti2;

	for ( i = 0; i < pla->getTerms(); i++ ) {
		matrix[i][i] = -1;
		ti1 = pla->getTermByIndex(i);
		for ( j = i+1; j < pla->getTerms(); j++ ) {
			ti2 = pla->getTermByIndex(j);
			k = (*ti1).intersectSize(*ti2);
			matrix[i][j] = k;
			matrix[j][i] = k;
		}
	}
}

void Ortho::AppendMatrixRow( const TermList::iterator ti )
/*
	Increases matrix by 1 and computes values for added term ti
*/
{
    vector<int> matrixrow;
	unsigned int i, j, k, l;
    list<Term>::iterator ti2;

	i = matrix.size();
	l = pla->getIndex(ti);
	if ( i + 1 != pla->getTerms() ) throw RuntimeException( "Ortho::AppendMatrixRow: Wrong matrix size" );
	matrixrow.resize(pla->getTerms());	
	matrix.push_back(matrixrow);
	for ( j = 0; j < i; j++ ) {
		matrix[j].push_back(-1);
		ti2 = pla->getTermByIndex(j);
		k = (*ti).intersectSize(*ti2);
		matrix[i][j] = k;
		matrix[j][i] = k;
	}
}

void Ortho::RecomputeMatrixRow( const TermList::iterator ti )
/*
	Recomputes values for term ti
*/
{
    unsigned int i, j, k;
    list<Term>::iterator ti2;

//	pla->getTerm(0);
	i = pla->getIndex(ti);
	for ( j = 0; j < pla->getTerms(); j++ ) 
		if ( j != i && matrix[i][j] != -1 ) {
			ti2 = pla->getTermByIndex(j);
			k = (*ti).intersectSize(*ti2);
			matrix[i][j] = k;
			matrix[j][i] = k;
		}
}

void Ortho::RemoveAbsorbed(void)
//	Removed absorbed terms from PLA
//  If IM is equal to some existing term, merges the term and deletes ti.
{
    TermList::iterator ti, ti1;

	if ( pla->getData()->size() < 2 ) return;
	ti = pla->getData()->begin();
	do {
        ti1 = ti;
        ti1++;
        while ( ti1 != pla->getData()->end() ) {
			if ( (*ti).contains( *ti1 ) ) ti1 = pla->DeleteTerm( ti1 );
			else if ( (*ti1).contains( *ti ) ) {
				ti = pla->DeleteTerm( ti );
				break;
			}
			else if ( (*ti).getCube() == (*ti1).getCube() ) {
				(*ti).Merge(*ti1);
				ti1 = pla->DeleteTerm( ti1 );
			} else ti1++;
        }
		ti++;
    } while ( ti != pla->getData()->end() );
}

bool Ortho::RemoveAbsorbed( const TermList::iterator ti )
//	Remove ti if absorbed
//  If IM is equal to some existing term, merges the term and deletes ti.
//  Returns true if it was removed
{
    TermList::iterator ti1;

    for ( ti1 = pla->getData()->begin(); ti1 != pla->getData()->end(); ti1++ ) 
		if ( ti1 != ti ) {
			if ( (*ti1).contains( *ti ) ) {
				pla->DeleteTerm( ti );
				return true;
			}
			if ( (*ti).getCube() == (*ti1).getCube() ) {
				(*ti1).Merge(*ti);
				pla->DeleteTerm( ti );
				return true;
			}
		}
	return false;
}

bool Ortho::TryToMerge( const TermList::iterator ti )
//	Tries to merge the term
//  Returns true if it was merged
{
    TermList::iterator ti1;

    for ( ti1 = pla->getData()->begin(); ti1 != pla->getData()->end(); ti1++ ) 
		if ( ti1 != ti ) {
			if ( (*ti).getCube() == (*ti1).getCube() ) {
				(*ti1).Merge(*ti);
				return true;
			}
		}
	return false;
}

bool Ortho::IsAbsorbed( const Term &t ) const
//	Returns true if t is absorbed
//  If IM is equal to some existing term, merges the term and returns true.
{
    TermList::iterator ti1;

	for ( ti1 = pla->getData()->begin(); ti1 != pla->getData()->end(); ti1++ ) {
		if ( t.getCube() == (*ti1).getCube() ) {
			(*ti1).Merge(t);
			return true;
		}
		else if ( (*ti1).contains(t) ) {
			return true;
		}
	}
	return false;
}

bool Ortho::WillBeAbsorbed( const TermList::iterator ti ) const
//	Returns true if t can be absorbed or merged
{
    TermList::iterator ti1;

	for ( ti1 = pla->getData()->begin(); ti1 != pla->getData()->end(); ti1++ ) 
		if ( ti != ti1 ) {
			if ( (*ti).getCube() == (*ti1).getCube() ) return true;
			else if ( (*ti1).contains((*ti)) ) 
				return true;
			
		}
	return false;
}

void Ortho::Orthogonalize2(void)
/*  Select term intersecting with most terms.
	Then select splitting variables, so that one term (both) will be absorbed.
	From it, splits by variable present in most intersected terms
*/
{
	Term newterm;
    list<Term>::iterator ti1, ti2, ti3, ti4;
    vector<int> candidates;
	vector<splitvar> candidates2;
	vector<int> matrixrow;
    unsigned int prom;
	short pol;
    unsigned int i, j;
    list<Term> *data;
	unsigned int cnt;
	unsigned int mx, cnt1, cnt0;
	unsigned int term;
	unsigned int val;
	bool absorbed1, absorbed2;		// are old/new terms absorbed?
	bool abs2;					// both terms are absorbed

	unsigned int tr1, tr2, tr3;
	tr1 = 0;
	tr2 = 0;
	tr3 = 0;

//	ofstream debugfile;
//	debugfile.open("ortho_debug.txt", std::ios::app);

	data = pla->getData();
				// first preprocessing
	RemoveAbsorbed();
	pla->MergeTerms();
	MoveOKTerms2();
//	pla->SavePLA("bad.pla");
//	ok_terms->SavePLA("ok.pla");
	GenerateMatrix();

	while ( pla->getTerms() != 0 ) {
		
			// find term conflicting with most terms
		mx = 0;
		candidates.clear();
		for ( i = 0; i < pla->getTerms(); i++ ) 
			if ( (*pla->getTermByIndex(i)).getDim() != 0 ) {
				cnt = 0;
				for ( j = 0; j < pla->getTerms(); j++ ) if ( matrix[i][j] != -1 ) cnt++;
				if ( cnt > mx ) {
					candidates.clear();
					candidates.push_back(i);
					mx = cnt;
				} else if ( cnt == mx && cnt != 0 )
					candidates.push_back(i);
			}
				// select one randomly
		if ( candidates.size() == 0 ) throw RuntimeException( "Ortho::Something's wrong - no intersection terms found" );
		term = candidates[rand() % candidates.size()];
		ti1 = pla->getTermByIndex(term);

				// select splitting variable, so that one term (both) will be absorbed
		abs2 = false;
		candidates2.clear();
		for ( i = 0; i < pla->inputs(); i++ ) 
			if ( (*ti1).isLit_dc(i) ) {
				(*ti1).setLit_1(i);
				if ( abs2 ) {	// now both have to be absorbed
					if ( WillBeAbsorbed(ti1) ) {
						(*ti1).setLit_0(i);
						if ( WillBeAbsorbed(ti1) ) {
							s_var.var = i;
							s_var.pol = 0;
							candidates2.push_back(s_var);
						}
					}
				} else {
					absorbed1 = WillBeAbsorbed(ti1);
					(*ti1).setLit_0(i);
					absorbed2 = WillBeAbsorbed(ti1);
					if ( absorbed1 && absorbed2 ) {
						abs2 = 1;
						candidates.clear();
						s_var.var = i;
						s_var.pol = 0;
						candidates2.push_back(s_var);
					} else if ( absorbed1 ) {
						s_var.var = i;
						s_var.pol = 1;
						candidates2.push_back(s_var);
					} else if ( absorbed2 ) {
						s_var.var = i;
						s_var.pol = -1;
						candidates2.push_back(s_var);
					}
				}
				(*ti1).setLit_dc(i);
			}

			// if there are no candidates, select the variable present in most intersected terms
//		debugfile << candidates2.size() << endl;
		if ( candidates2.empty() ) {
			mx = 0;
			for ( i = 0; i < pla->inputs(); i++ ) {
				if ( (*ti1).isLit_dc(i) ) {
					cnt0 = 0;
					cnt1 = 0;
					for ( ti2 = data->begin(); ti2 != data->end(); ++ti2 ) {
						if ( matrix[term][pla->getIndex(ti2)] != -1 ) {
							if ( (*ti2).isLit_0(i) ) cnt1++;
							else if ( (*ti2).isLit_1(i) ) cnt0++;
						}
					}
					val = max( cnt0, cnt1 );
					if ( mx < val ) {
						candidates2.clear();
						s_var.var = i;
						s_var.pol = 0;
						candidates2.push_back(s_var);
						mx = val;
					} else if ( mx == val ) {
						s_var.var = i;
						s_var.pol = 0;
						candidates2.push_back(s_var);
					}
				}
			}
		}
				// select one randomly
		if ( candidates2.empty() ) throw RuntimeException( "Ortho::Something's wrong - no intersection terms found" );
		i = rand() % candidates2.size();
		prom = candidates2[i].var;
		pol = candidates2[i].pol;
		
				// split the term
		newterm = *ti1;
		(*ti1).setLit_1(prom);
		newterm.setLit_0(prom);
		absorbed1 = RemoveAbsorbed( ti1 );
		absorbed2 = IsAbsorbed( newterm );
		if ( !absorbed2 ) ti2 = pla->AppendTerm( newterm );

			// recompute everything
		tr1++;
		if ( !absorbed1 && absorbed2 ) {
			RecomputeMatrix();
			tr2++;
		}
		// else if ( !absorbed1 && !absorbed2 ) AppendMatrixRow( ti2 );   // this only slows it down
		else {
			GenerateMatrix();
			tr3++;
		}
		i = pla->getTerms();
		MoveOKTerms();
		if ( pla->getTerms() != i ) GenerateMatrix();
//		pla->SavePLA("bad.pla");
//		ok_terms->SavePLA("ok.pla");
	}
	if ( !IsOrthogonal(ok_terms) ) throw RuntimeException( "Ortho::Something's wrong - resulting PLA is not orthogonal" );
//	debugfile << endl;
//	debugfile.close();
//	cout << "   All/Recomputed/Generated " << tr1 << " " << tr2 << " " << tr3 << endl;
}

void Ortho::Orthogonalize2a(void)
/*  Select term intersecting with most terms.
	Then select splitting variables, so that one term (both) will be absorbed.
	From it, splits by variable present in most intersected terms
	No intersection matrix is used.

	It is slower than Orthogonalize2!
*/
{
	Term newterm;
    list<Term>::iterator ti1, ti2, ti3, ti4;
    vector< list<Term>::iterator > candidates;	// candidate terms for the selection
	vector<splitvar> candidates2;	// candidate variables for the selection
    unsigned int prom;
	short pol;
    unsigned int i;
    list<Term> *data;
	unsigned int cnt;
	unsigned int mx, cnt1, cnt0;
	unsigned int val;
	bool absorbed1, absorbed2;		// are old/new terms absorbed?
	bool abs2;					// both terms are absorbed

//	ofstream debugfile;
//	debugfile.open("ortho_debug.txt", std::ios::app);

	data = pla->getData();
				// first preprocessing
	RemoveAbsorbed();
	pla->MergeTerms();
	MoveOKTerms2();
//	pla->SavePLA("bad.pla");
//	ok_terms->SavePLA("ok.pla");

	 while ( pla->getTerms() != 0 ) {
			// find term conflicting with most terms
		mx = 0;
		candidates.clear();
		for ( ti1 = pla->getData()->begin(); ti1 != pla->getData()->end(); ti1++ ) 
			if ( (*ti1).getDim() != 0 ) {
				cnt = 0;
				ti2 = ti1;
				ti2++;
				for ( ti2 = pla->getData()->begin(); ti2 != pla->getData()->end(); ti2++ )
					if ( ti1 != ti2 ) 
						if ( (*ti1).intersects(*ti2) ) cnt++;
				if ( cnt > mx ) {
					candidates.clear();
					candidates.push_back(ti1);
					mx = cnt;
				} else if ( cnt == mx && cnt != 0 )
					candidates.push_back(ti1);
			}
				// select one randomly
		if ( candidates.size() == 0 ) throw RuntimeException( "Ortho::Something's wrong - no intersection terms found" );
		ti1 = candidates[rand() % candidates.size()];

				// select splitting variable, so that one term (both) will be absorbed
		abs2 = false;
		candidates2.clear();
		for ( i = 0; i < pla->inputs(); i++ ) 
			if ( (*ti1).isLit_dc(i) ) {
				(*ti1).setLit_1(i);
				if ( abs2 ) {	// now both have to be absorbed
					if ( WillBeAbsorbed(ti1) ) {
						(*ti1).setLit_0(i);
						if ( WillBeAbsorbed(ti1) ) {
							s_var.var = i;
							s_var.pol = 0;
							candidates2.push_back(s_var);
						}
					}
				} else {
					absorbed1 = WillBeAbsorbed(ti1);
					(*ti1).setLit_0(i);
					absorbed2 = WillBeAbsorbed(ti1);
					if ( absorbed1 && absorbed2 ) {
						abs2 = 1;
						candidates.clear();
						s_var.var = i;
						s_var.pol = 0;
						candidates2.push_back(s_var);
					} else if ( absorbed1 ) {
						s_var.var = i;
						s_var.pol = 1;
						candidates2.push_back(s_var);
					} else if ( absorbed2 ) {
						s_var.var = i;
						s_var.pol = -1;
						candidates2.push_back(s_var);
					}
				}
				(*ti1).setLit_dc(i);
			}

			// if there are no candidates, select the variable present in most intersected terms
//		debugfile << candidates2.size() << endl;
		if ( candidates2.empty() ) {
			mx = 0;
			for ( i = 0; i < pla->inputs(); i++ ) {
				if ( (*ti1).isLit_dc(i)) {
					cnt0 = 0;
					cnt1 = 0;
					for ( ti2 = data->begin(); ti2 != data->end(); ++ti2 ) {
						if ( (*ti1).intersects(*ti2) ) {
							if ( (*ti2).isLit_0(i) ) cnt1++;
							else if ( (*ti2).isLit_1(i) ) cnt0++;
						}
					}
					val = max( cnt0, cnt1 );
					if ( mx < val ) {
						candidates2.clear();
						s_var.var = i;
						s_var.pol = 0;
						candidates2.push_back(s_var);
						mx = val;
					} else if ( mx == val ) {
						s_var.var = i;
						s_var.pol = 0;
						candidates2.push_back(s_var);
					}
				}
			}
		}
				// select one randomly
		if ( candidates2.empty() ) throw RuntimeException( "Ortho::Something's wrong - no intersection terms found" );
		i = rand() % candidates2.size();
		prom = candidates2[i].var;
		pol = candidates2[i].pol;
		
				// split the term
		newterm = *ti1;
		(*ti1).setLit_1(prom );
		newterm.setLit_0(prom);
		absorbed1 = RemoveAbsorbed( ti1 );
		absorbed2 = IsAbsorbed( newterm );
		if ( !absorbed2 ) ti2 = pla->AppendTerm( newterm );

			// recompute everything
		MoveOKTerms2();
//		pla->SavePLA("bad.pla");
//		ok_terms->SavePLA("ok.pla");
	}
//	if ( !IsOrthogonal(ok_terms) ) throw RuntimeException( "Ortho::Something's wrong - resulting PLA is not orthogonal" );
//	debugfile << endl;
//	debugfile.close();
}

void Ortho::Orthogonalize3(void)
/*  1. find a term & variable to split, so that one (or both) split terms will be absorbed
	2. from these candidates (if any) find term conflicting with most terms
	3. from these candidates select the variable present in most intersected terms
*/
{
	Term newterm;
    list<Term>::iterator ti1, ti2, ti3, ti4;
    vector<int> candidates;
	vector<splitvar> candidates2;
	vector<splitterm> candidates3;
	vector<splitterm>::iterator cand3_it;
	vector<int> matrixrow;
    unsigned int prom;
	short pol;
    unsigned int i, j;
    list<Term> *data;
	unsigned int cnt;
	unsigned int mx, cnt1, cnt0;
	unsigned int term;
	unsigned int val;
	bool absorbed1, absorbed2;		// are old/new terms absorbed?
	bool abs2;					// both terms are absorbed
	set<unsigned int> termset;
	set<unsigned int> varset;
	set<unsigned int>::iterator varset_it;

	unsigned int tr1, tr2, tr3;
	tr1 = 0;
	tr2 = 0;
	tr3 = 0;

//	ofstream debugfile;
//	debugfile.open("ortho_debug.txt", std::ios::app);

	data = pla->getData();
				// first preprocessing
	RemoveAbsorbed();
	pla->MergeTerms();
	MoveOKTerms2();
//	pla->SavePLA("bad.pla");
//	ok_terms->SavePLA("ok.pla");
	GenerateMatrix();

	while ( pla->getTerms() != 0 ) {

			// find a term & variable to split, so that one (or both) split terms will be absorbed
		abs2 = false;
		candidates3.clear();
		termset.clear();
		for ( ti1 = pla->getData()->begin(); ti1 != pla->getData()->end(); ti1++ ) 
			if ( (*ti1).getDim() != 0 )
				for ( i = 0; i < pla->inputs(); i++ ) 
					if ( (*ti1).isLit_dc(i) ) {
						(*ti1).setLit_1(i);
						if ( abs2 ) {	// now both have to be absorbed to be accepted
							if ( WillBeAbsorbed(ti1) ) {
								(*ti1).setLit_0(i);
								if ( WillBeAbsorbed(ti1) ) {
									s_term.ti = ti1;
									s_term.var = i;
									s_term.pol = 0;
									candidates3.push_back(s_term);
									termset.insert(pla->getIndex(ti1));
								}
							}
						} else {
							absorbed1 = WillBeAbsorbed(ti1);
							(*ti1).setLit_0(i);
							absorbed2 = WillBeAbsorbed(ti1);
							if ( absorbed1 && absorbed2 ) {
								abs2 = true;
								candidates3.clear();
								s_term.ti = ti1;
								s_term.var = i;
								s_term.pol = 0;
								termset.insert(pla->getIndex(ti1));
								candidates3.push_back(s_term);
							} else if ( absorbed1 ) {
								s_term.ti = ti1;
								s_term.var = i;
								s_term.pol = 1;
								termset.insert(pla->getIndex(ti1));
								candidates3.push_back(s_term);
							} else if ( absorbed2 ) {
								s_term.ti = ti1;
								s_term.var = i;
								s_term.pol = -1;
								termset.insert(pla->getIndex(ti1));
								candidates3.push_back(s_term);
							}
						}
						(*ti1).setLit_dc(i);
					}

					// from these candidates (if any) find term conflicting with most terms
		mx = 0;
		candidates.clear();
		for ( i = 0; i < pla->getTerms(); i++ ) 
			if ( (*pla->getTermByIndex(i)).getDim() != 0 && (termset.empty() || termset.find(i) != termset.end()) ) {
				cnt = 0;
				for ( j = 0; j < pla->getTerms(); j++ ) if ( matrix[i][j] != -1 ) cnt++;
				if ( cnt > mx ) {
					candidates.clear();
					candidates.push_back(i);
					mx = cnt;
				} else if ( cnt == mx && cnt != 0 )
					candidates.push_back(i);
			}
				// select one randomly
		if ( candidates.size() == 0 ) throw RuntimeException( "Orthogonalize3::Something's wrong - no intersection terms found" );
		term = candidates[rand() % candidates.size()];
		ti1 = pla->getTermByIndex(term);

			// create a candidate set for splitting variables
		varset.clear();
		if ( candidates3.empty() ) {
			for ( i = 0; i < pla->inputs(); i++ ) 
				if ( (*ti1).isLit_dc(i) ) varset.insert(i);
		} else {
			for ( cand3_it = candidates3.begin(); cand3_it != candidates3.end(); cand3_it++ )
				if ( (*cand3_it).ti == ti1 ) varset.insert( (*cand3_it).var );
		}

			// select the variable present in most intersected terms
		if ( varset.size() > 1 ) {
			mx = 0;
			candidates2.clear();
			for ( varset_it = varset.begin();  varset_it != varset.end(); varset_it++ ) {
				i = *varset_it;
				cnt0 = 0;
				cnt1 = 0;
				for ( ti2 = data->begin(); ti2 != data->end(); ++ti2 ) {
					if ( matrix[term][pla->getIndex(ti2)] != -1 ) {
						if ( (*ti2).isLit_0(i) ) cnt1++;
						else if ( (*ti2).isLit_1(i) ) cnt0++;
					}
				}
				val = max( cnt0, cnt1 );
				if ( mx < val ) {
					candidates2.clear();
					s_var.var = i;
					s_var.pol = 0;
					candidates2.push_back(s_var);
					mx = val;
				} else if ( mx == val ) {
					s_var.var = i;
					s_var.pol = 0;
					candidates2.push_back(s_var);
				}
			}
					// select one randomly
//			pla->SavePLA("bad.pla");
			if ( candidates2.empty() ) throw RuntimeException( "Orthogonalize3::Something's wrong - no intersection terms found" );
			i = rand() % candidates2.size();
			prom = candidates2[i].var;
			pol = candidates2[i].pol;
		} else if (varset.size() == 1 ) {
			prom = *(varset.begin());
			pol = 0;
		} else {						// varset empty. No splitting variable helps. Select one randomly
			i = rand() % pla->inputs() + 1;
			prom = 0;
			do {
				if ( (*ti1).isLit_dc(prom) ) i--;
				else {
					prom++;
					prom %= pla->inputs();
				}
			} while ( i > 0 );
		}


//		debugfile << candidates3.size() << endl;
		
				// split the term
		if ( !(*ti1).isLit_dc(prom) ) throw RuntimeException( "Orthogonalize3::Something's wrong - bad splitting variable" );
		newterm = *ti1;
		(*ti1).setLit_1(prom);
		newterm.setLit_0(prom);
		absorbed1 = RemoveAbsorbed( ti1 );
		absorbed2 = IsAbsorbed( newterm );
		if ( !absorbed2 ) ti2 = pla->AppendTerm( newterm );

			// recompute matrix
		tr1++;

		if ( !absorbed1 && absorbed2 ) {
			RecomputeMatrix();
			tr2++;
		}
		// else if ( !absorbed1 && !absorbed2 ) AppendMatrixRow( ti2 );   // this only slows it down
		else {
			GenerateMatrix();
			tr3++;
		}
		i = pla->getTerms();
		MoveOKTerms();
		if ( pla->getTerms() != i && pla->getTerms() != 0 ) GenerateMatrix();
		// pla->SavePLA("bad.pla");
		// ok_terms->SavePLA("ok.pla");
	}
	if ( !IsOrthogonal(ok_terms) ) throw RuntimeException( "Orthogonalize3::Something's wrong - resulting PLA is not orthogonal" );
//	debugfile << endl;
//	debugfile.close();
//	cout << "   All/Recomputed/Generated: " << tr1 << " " << tr2 << " " << tr3 << endl;
}

void Ortho::Orthogonalize4(void)
/*  1. find a term & variable to split, so that one (or both) split terms will be absorbed.
	2. from these, choose terms&variables, where splitting variables are present in most intersected terms
*/
{
	Term newterm;
    list<Term>::iterator ti1, ti2, ti3, ti4;
    vector<int> candidates;
	vector<splitterm> candidates3;
	vector<splitterm> candidates2;
	vector<splitterm>::iterator cand3_it;
	vector<int> matrixrow;
    unsigned int prom;
    unsigned int i;
    list<Term> *data;
	unsigned int mx, cnt1, cnt0;
	unsigned int val;
	bool absorbed1, absorbed2;		// are old/new terms absorbed?
	bool abs2;					// both terms are absorbed

	unsigned int tr1, tr2, tr3;
	tr1 = 0;
	tr2 = 0;
	tr3 = 0;

//	ofstream debugfile;
//	debugfile.open("ortho_debug.txt", std::ios::app);

	data = pla->getData();
				// first preprocessing
	RemoveAbsorbed();
	pla->MergeTerms();
	MoveOKTerms2();
//	pla->SavePLA("bad.pla");
//	ok_terms->SavePLA("ok.pla");
	GenerateMatrix();

	while ( pla->getTerms() != 0 ) {
			// find a term & variable to split, so that one (or both) split terms will be absorbed
		abs2 = false;
		candidates3.clear();
		for ( ti1 = pla->getData()->begin(); ti1 != pla->getData()->end(); ti1++ ) 
			if ( (*ti1).getDim() != 0 )
				for ( i = 0; i < pla->inputs(); i++ ) 
					if ( (*ti1).isLit_dc(i) ) {
						(*ti1).setLit_1(i);
						if ( abs2 ) {	// now both have to be absorbed to be accepted
							if ( WillBeAbsorbed(ti1) ) {
								(*ti1).setLit_0(i);
								if ( WillBeAbsorbed(ti1) ) {
									s_term.ti = ti1;
									s_term.var = i;
									s_term.pol = 0;
									candidates3.push_back(s_term);
								}
							}
						} else {
							absorbed1 = WillBeAbsorbed(ti1);
							(*ti1).setLit_0(i);
							absorbed2 = WillBeAbsorbed(ti1);
							if ( absorbed1 && absorbed2 ) {
								abs2 = true;
								candidates3.clear();
								s_term.ti = ti1;
								s_term.var = i;
								s_term.pol = 0;
								candidates3.push_back(s_term);
							} else if ( absorbed1 ) {
								s_term.ti = ti1;
								s_term.var = i;
								s_term.pol = 1;
								candidates3.push_back(s_term);
							} else if ( absorbed2 ) {
								s_term.ti = ti1;
								s_term.var = i;
								s_term.pol = -1;
								candidates3.push_back(s_term);
							}
						}
						(*ti1).setLit_dc(i);
					}
		if ( candidates3.empty() ) 		// no candidates, generate all (terms x DC-vars)	
			for ( ti1 = pla->getData()->begin(); ti1 != pla->getData()->end(); ti1++ ) 
				if ( (*ti1).getDim() != 0 )
					for ( i = 0; i < pla->inputs(); i++ )
						if ( (*ti1).isLit_dc(i) ) {
							s_term.ti = ti1;
							s_term.var = i;
							s_term.pol = 0;
							candidates3.push_back(s_term);
						}
		if ( candidates3.empty() ) throw RuntimeException( "Orthogonalize::Ortho4: Something's wrong. No splitting variables" );

		// select the variable (+term) present in most intersected terms
			
		mx = 0;
		candidates2.clear();
		for ( cand3_it = candidates3.begin();  cand3_it != candidates3.end(); ++cand3_it ) {
			cnt0 = 0;
			cnt1 = 0;
			for ( ti2 = data->begin(); ti2 != data->end(); ++ti2 ) {
				if ( matrix[pla->getIndex((*cand3_it).ti)][pla->getIndex(ti2)] != -1 ) {
					if ( (*ti2).isLit_0((*cand3_it).var) ) cnt1++;
					else if ( (*ti2).isLit_1((*cand3_it).var) ) cnt0++;
				}
			}
			val = max( cnt0, cnt1 );
			if ( mx < val ) {
				candidates2.clear();
				candidates2.push_back(*cand3_it);
				mx = val;
			} else if ( mx == val && mx != 0 )
				candidates2.push_back(*cand3_it);
		}
		if ( candidates2.empty() ) {		// splitting these terms by its variables will not help. Do it once more with all terms.
			for ( ti1 = pla->getData()->begin(); ti1 != pla->getData()->end(); ti1++ ) 
				for ( i = 0; i < pla->inputs(); i++ )
					if ( (*ti1).isLit_dc(i) ) {
						s_term.ti = ti1;
						s_term.var = i;
						s_term.pol = 0;
						candidates3.push_back(s_term);
					}
			for ( cand3_it = candidates3.begin();  cand3_it != candidates3.end(); ++cand3_it ) {
				cnt0 = 0;
				cnt1 = 0;
				for ( ti2 = data->begin(); ti2 != data->end(); ++ti2 ) {
					if ( matrix[pla->getIndex((*cand3_it).ti)][pla->getIndex(ti2)] != -1 ) {
						if ( (*ti2).isLit_0((*cand3_it).var) ) cnt1++;
						else if ( (*ti2).isLit_1((*cand3_it).var) ) cnt0++;
					}
				}
				val = max( cnt0, cnt1 );
				if ( mx < val ) {
					candidates2.clear();
					candidates2.push_back(*cand3_it);
					mx = val;
				} else if ( mx == val && mx != 0 )
					candidates2.push_back(*cand3_it);
			}
		}
		if ( candidates2.empty() ) throw RuntimeException( "Orthogonalize::Ortho4: Something's wrong. No splitting variables" );
	
		// select one randomly

		i = rand() % candidates2.size();
		prom = candidates2[i].var;
		ti1 = candidates2[i].ti;

		// split it

		if ( !(*ti1).isLit_dc(prom) ) throw RuntimeException( "Orthogonalize4::Something's wrong - bad splitting variable" );
		newterm = *ti1;
		(*ti1).setLit_1(prom);
		newterm.setLit_0(prom);
		absorbed1 = RemoveAbsorbed( ti1 );
		absorbed2 = IsAbsorbed( newterm );
		if ( !absorbed2 ) ti2 = pla->AppendTerm( newterm );

			// recompute matrix
		tr1++;

		if ( !absorbed1 && absorbed2 ) {
			RecomputeMatrix();
			tr2++;
		}
		// else if ( !absorbed1 && !absorbed2 ) AppendMatrixRow( ti2 );   // this only slows it down
		else {
			GenerateMatrix();
			tr3++;
		}
		i = pla->getTerms();
		MoveOKTerms();
		if ( pla->getTerms() != i && pla->getTerms() != 0 ) GenerateMatrix();
//		pla->SavePLA("bad.pla");
//		ok_terms->SavePLA("ok.pla");
	
	}

	if ( !IsOrthogonal(ok_terms) ) throw RuntimeException( "Orthogonalize4::Something's wrong - resulting PLA is not orthogonal" );
}

void Ortho::Orthogonalize5(void)
/*  1. find a term & variable to split, so that one (or both) split terms will be absorbed and, by this splitting,
		most intersected terms won't be intersected

*/
{
	Term newterm;
    list<Term>::iterator ti1, ti2, ti3, ti4;
    vector<int> candidates;
	vector<splitterm> candidates3;
	vector<splitterm> candidates2;
	vector<splitterm>::iterator cand3_it;
	vector<int> matrixrow;
    unsigned int prom;
	short pol;
    unsigned int i, j;
    list<Term> *data;
	unsigned int mx, cnt1, cnt0;
	unsigned int term;
	unsigned int val;
	bool absorbed1, absorbed2;		// are old/new terms absorbed?
	unsigned int orig_intersects, intersects;	// numbers of intersected terms
	bool consensus;				// consensus detected

	unsigned int tr1, tr2, tr3;
	tr1 = 0;
	tr2 = 0;
	tr3 = 0;

//	ofstream debugfile;
//	debugfile.open("ortho_debug.txt", std::ios::app);

	data = pla->getData();
				// first preprocessing
	RemoveAbsorbed();
	pla->MergeTerms();
	MoveOKTerms2();
//	pla->SavePLA("bad.pla");
//	ok_terms->SavePLA("ok.pla");
	GenerateMatrix();

	while ( pla->getTerms() != 0 ) {
			// find a term & variable to split, so that one (or both) split terms will be absorbed
		candidates3.clear();
		ti1 = data->begin();
		mx = 0;
		do {
			consensus = false;
			if ( (*ti1).getDim() != 0 ) {
				orig_intersects = 0;		// compute # of intersecting terms
				for ( i = 0; i < pla->getTerms(); i++ )
					if ( matrix[pla->getIndex(ti1)][i] != -1 ) orig_intersects++;
				if ( orig_intersects >= mx ) {		// otherwise this term cannot compete
					for ( i = 0; i < pla->inputs(); i++ ) 
						if ( (*ti1).isLit_dc(i) ) {
							(*ti1).setLit_1(i);
							absorbed1 = WillBeAbsorbed(ti1);
							(*ti1).setLit_0(i);
							absorbed2 = WillBeAbsorbed(ti1);
							if ( absorbed1 && absorbed2 ) {		// it is the consenzus or could be merged. Remove it.
								// (*ti1).setLit_0(i);		// is not necessary here. It was set above
								TryToMerge(ti1);
								(*ti1).setLit_1(i);
								TryToMerge(ti1);
								cand3_it = candidates3.begin();
								while ( cand3_it != candidates3.end() ) {			// delete the term, if it was in candidates
									if ( (*cand3_it).ti == ti1 ) cand3_it = candidates3.erase(cand3_it);
									else cand3_it++;
								}
								ti1 = pla->DeleteTerm(ti1);
								GenerateMatrix();
								j = pla->getTerms();
								MoveOKTerms();
								if ( pla->getTerms() != j && pla->getTerms() != 0 ) {
									GenerateMatrix();
									ti1 = data->begin();
									candidates3.clear();
								}
								//	ti1 = data->begin();		// this somehow produces worse results
								//	candidates3.clear();
								consensus = true;
								break;		// exit the for loop
							} else if ( absorbed1 ) {
								// (*ti1).setLit( i, '0' );		// is not necessary here. It was set above
								intersects = 0;			// compute # of intersections
								for ( j = 0; j < pla->getTerms(); j++ )
									if ( matrix[j][pla->getIndex(ti1)] != -1 )
										if ( (*ti1).intersects( *(pla->getTermByIndex(j)) )) intersects++;
								if ( orig_intersects - intersects == mx && mx != 0 ) {			
									s_term.ti = ti1;
									s_term.var = i;
									s_term.pol = 0;
									candidates3.push_back(s_term);
								} else if ( orig_intersects - intersects > mx ) {			
									s_term.ti = ti1;
									s_term.var = i;
									s_term.pol = 0;
									candidates3.clear();
									candidates3.push_back(s_term);
									mx = orig_intersects - intersects;
								}
							} else if ( absorbed2 ) {
								(*ti1).setLit_1(i);
								intersects = 0;			// compute # of intersections
								for ( j = 0; j < pla->getTerms(); j++ )
									if ( matrix[j][pla->getIndex(ti1)] != -1 )
										if ( (*ti1).intersects( *(pla->getTermByIndex(j)) )) intersects++;
								if ( orig_intersects - intersects == mx  && mx != 0 ) {			
									s_term.ti = ti1;
									s_term.var = i;
									s_term.pol = 1;
									candidates3.push_back(s_term);
								} else if ( orig_intersects - intersects > mx ) {			
									s_term.ti = ti1;
									s_term.var = i;
									s_term.pol = 1;
									candidates3.clear();
									candidates3.push_back(s_term);
									mx = orig_intersects - intersects;
								}
							}
							(*ti1).setLit_dc(i);
						}
				}
			}
			if (!consensus) ti1++;
		} while ( pla->getTerms() != 0 && ti1 != data->end() );	  
		if ( pla->getTerms() == 0 ) break;	// consensus eliminated all terms
		if ( !candidates3.empty() ) {				// one is absorbed
					// select one randomly
			i = rand() % candidates3.size();
			prom = candidates3[i].var;
			ti1 = candidates3[i].ti;
			pol = candidates3[i].pol;
			if ( pol == 0 ) (*ti1).setLit_1(prom);  // try to merge the absorbed one
			else (*ti1).setLit_0(prom);
			TryToMerge(ti1);
			if ( pol == 0 ) (*ti1).setLit_0(prom);
			else (*ti1).setLit_1(prom);
			RecomputeMatrix();
		} else
		{									// no chance for absobption
				// find term intersecting with most other terms
			mx = 0;
			candidates.clear();
			for ( ti1 = data->begin(); ti1 != data->end(); ti1++ ) 
				if ( (*ti1).getDim() != 0 ) {
					orig_intersects = 0;
					for ( i = 0; i < pla->getTerms(); i++ )
						if ( matrix[pla->getIndex(ti1)][i] != -1 ) orig_intersects++;
					if ( orig_intersects == mx  && mx != 0 ) {
						candidates.push_back(pla->getIndex(ti1));
					} else if ( orig_intersects > mx ) {			
						candidates.clear();
						candidates.push_back(pla->getIndex(ti1));
						mx = orig_intersects;
					}
				}
				// select one randomly
			if ( candidates.empty() ) throw RuntimeException( "Orthogonalize::Ortho5: Something's wrong. No intersecting term." );
			term = candidates[rand() % candidates.size()];
			ti1 = pla->getTermByIndex(term);

				// select the variable present in most intersected terms
			mx = 0;
			candidates.clear();
			for ( i = 0; i < pla->inputs(); i++ ) 
				if ( (*ti1).isLit_dc(i) ) {
					cnt0 = 0;
					cnt1 = 0;
					for ( ti2 = data->begin(); ti2 != data->end(); ++ti2 ) {
						if ( matrix[term][pla->getIndex(ti2)] != -1 ) {
							if ( (*ti2).isLit_1(i) ) cnt1++;
							else if ( (*ti2).isLit_0(i) ) cnt0++;
						}
					}
					val = max( cnt0, cnt1 );
					if ( mx < val ) {
						candidates.clear();
						candidates.push_back(i);
						mx = val;
					} else if ( mx == val )
						candidates.push_back(i);
				}
			if ( candidates.empty() ) throw RuntimeException( "Orthogonalize::Ortho5: Something's wrong. No splitting variables" );
				// select one randomly
			prom = candidates[rand() % candidates.size()];

			newterm = *ti1;
			(*ti1).setLit_1(prom);
			newterm.setLit_0(prom);
			pla->AppendTerm( newterm );
			GenerateMatrix();
		}
		i = pla->getTerms();
		RemoveAbsorbed();
		if ( pla->getTerms() != i && pla->getTerms() != 0 ) {
			GenerateMatrix();
			i = pla->getTerms();
		}
		MoveOKTerms();
		if ( pla->getTerms() != i && pla->getTerms() != 0 ) GenerateMatrix();

		pla->SavePLA("bad.pla");
		ok_terms->SavePLA("ok.pla");
	}

	if ( !IsOrthogonal(ok_terms) ) throw RuntimeException( "Orthogonalize5::Something's wrong - resulting PLA is not orthogonal" );
}

// ================== Hoger ======================

void Ortho::SizeSortInc(void)
/*
Term size sort Incremmentally
*/
{ 
    PLA *list_result; // result PLA
    list_result = new PLA(pla);	

    list<Term> res; // result of operations    
	TermList::const_iterator t1, t2;
    
   	Term useterm; // working term 
    int counter = 0; // counter of terms
    int t_size = 0; // counter of terms
    
    int * array_dc; // array of dc values
    array_dc = (int*)malloc(90000*sizeof(int));
      
    /* find term size */
    t1 = pla->getData()->begin();
    useterm = *t1;
    t_size = useterm.inputs(); // term size
    
    /* counter dc for terms */ 
    for( t1 = pla->getData()->begin(); t1 != pla->getData()->end(); t1++ ) {
          useterm = *t1;
          array_dc[counter] = 0;
          
          /* count dc for invidual term*/ 
          for ( int j = 0; j <= t_size; j++ ) {
              if (useterm.isLit_dc(j)) { 
                 array_dc[counter]++;
              }   
          }
          counter++;
     }           

    int pointer_counter = 0; // pointer counter position
    
    /* create new sort pla */
    for ( int j = 0; j <= t_size; j++ ) {
      
         counter = 0;

         for( t2 = pla->getData()->begin(); t2 != pla->getData()->end(); t2++ ) {
               useterm = *t2;          
                   
               /* compare dc with grade term */      
               if (j == array_dc[counter]) {
                    /* save term */
                    res.push_back(useterm);
                    
                    /* change term order position */
                    pointer_counter++;
               }     
               counter++;
         }
    }                
     
   list_result->AppendTermList(res); // save result to termlist       	     
   delete pla;
   pla = list_result; // sort result

   free(array_dc); // deallocate memory (array of dc values)     
  
}

void Ortho::SizeSortDec(void)
/*
Term size sort Decremmentally
*/
{ 
    PLA *list_result; // result PLA
    list_result = new PLA(pla);	

    list<Term> res; // result of operations    
	TermList::const_iterator t1, t2;
    
   	Term useterm; // working term 
    int counter = 0; // counter of terms
    int t_size = 0; // counter of terms
    
    int * array_dc; // array of dc values
    array_dc = (int*)malloc(90000*sizeof(int));
      
    /* find term size */
    t1 = pla->getData()->begin();
    useterm = *t1;
    t_size = useterm.inputs(); // term size
    
    /* counter dc for terms */ 
    for( t1 = pla->getData()->begin(); t1 != pla->getData()->end(); t1++ ) {
          useterm = *t1;
          array_dc[counter] = 0;
          
          /* count dc for invidual term*/ 
          for ( int j = 0; j <= t_size; j++ ) {
              if ( useterm.isLit_dc(j) ) { 
                 array_dc[counter]++;
              }   
          }
          counter++;
     }           


    int pointer_counter = 0; // pointer counter position
    
    /* create new sort pla */
    for ( int j = t_size; 0 <= j; j-- ) {
      
         counter = 0;

         for( t2 = pla->getData()->begin(); t2 != pla->getData()->end(); t2++ ) {
               useterm = *t2;          
                   
               /* compare dc with grade term */      
               if (j == array_dc[counter]) {
                    /* save term */
                    res.push_back(useterm);
                    
                    /* change term order position */
                    pointer_counter++;
               }     
               counter++;
         }
    }                
     
   list_result->AppendTermList(res); // save result to termlist       	     
   delete pla;
   pla = list_result; // sort result

   free(array_dc); // deallocate memory (array of dc values)   
}

void Ortho::IntersectionSortInc(void)
/*
Term Intersect Sort incrementally
*/
{ 
    PLA *list_result; // result PLA
    list_result = new PLA(pla);	

    list<Term> res; // result of operations    
	TermList::const_iterator t1, t2;
    
   	Term useterm; // working term 
    int counter = 0; // counter of terms
    int i_max = 0; // max value
    
    int * array_ints; // array of dc values
    array_ints = (int*)malloc(90000*sizeof(int));
          
     /* counter intersects */  
    for( t1 = pla->getData()->begin(); t1 != pla->getData()->end(); t1++ ) {
     
       array_ints[counter] = 0;
       useterm = *t1;
          
       for( t2 = pla->getData()->begin(); t2 != pla->getData()->end(); t2++ ) {
         
          /* counts ints for invidual term*/ 
          if(t1 != t2) {
             if(useterm.intersects(*t2)) {
                 array_ints[counter]++;
                
                 if(i_max < array_ints[counter]) 
                    i_max = array_ints[counter];
             }
          } 
       }
       counter++;
    }           

    int pointer_counter = 0; // pointer counter position
    
    /* create new sort pla */
    for ( int j = 0; j <= i_max; j++ ) {
      
         counter = 0;
         for( t2 = pla->getData()->begin(); t2 != pla->getData()->end(); t2++ ) {
               useterm = *t2;          
                   
               /* compare ints with grade term */      
               if (j == array_ints[counter]) {
                    /* save term */
                    res.push_back(useterm);
                    
                    /* change term order position */
                    pointer_counter++;
               }     
               counter++;
         }
    }                
     
   list_result->AppendTermList(res); // save result to termlist       	     
   delete pla;
   pla = list_result; // sort result

   free(array_ints); // deallocate memory (array of ints values) 
}

void Ortho::IntersectionSortDec(void)
/*
Term Intersect Sort decrementally
*/
{ 
    PLA *list_result; // result PLA
    list_result = new PLA(pla);	

    list<Term> res; // result of operations    
	TermList::const_iterator t1, t2;
    
   	Term useterm; // working term 
    int counter = 0; // counter of terms
    int i_max = 0; // max value
    
    int * array_ints; // array of dc values
    array_ints = (int*)malloc(90000*sizeof(int));
         
    /* counter intersects */ 
    for( t1 = pla->getData()->begin(); t1 != pla->getData()->end(); t1++ ) {
     
       array_ints[counter] = 0;
       useterm = *t1;
          
       for( t2 = pla->getData()->begin(); t2 != pla->getData()->end(); t2++ ) {
         
          /* counts ints for invidual term*/ 
          if(t1 != t2) {
             if(useterm.intersects(*t2)) {
                 array_ints[counter]++;

                 if(i_max < array_ints[counter]) 
                    i_max = array_ints[counter];
             }
          } 
       }    
       counter++;
    }           

    int pointer_counter = 0; // pointer counter position
    
    /* create new sort pla */
    for ( int j = i_max; 0 <= j; j-- ) {
      
         counter = 0;
         for( t2 = pla->getData()->begin(); t2 != pla->getData()->end(); t2++ ) {
               useterm = *t2;          
                   
               /* compare ints with grade term */      
               if (j == array_ints[counter]) {
                    /* save term */
                    res.push_back(useterm);
                    
                    /* change term order position */
                    pointer_counter++;
               }     
               counter++;
         }
    }                
     
   list_result->AppendTermList(res); // save result to termlist       	     
   delete pla;
   pla = list_result; // sort result

   free(array_ints); // deallocate memory (array of ints values) 
}

void Ortho::SortByMergeInc(void)
/*
Term MergeNeighbors Sort incrementally
*/
{ 
    PLA use;
    PLA *list_result; // result PLA
    
    list_result = new PLA(pla);	
    
    list<Term> res; // result of operations    
	TermList::const_iterator t1, t2, t3;

	Term newterm;    
   	Term useterm; // working term 
 
    int t_sizeIn = 0; // term output size
    int t_sizeOu = 0; // term output size
      
    /* find term input and output size */
    t1 = pla->getData()->begin();
    useterm = *t1;
    t_sizeIn = useterm.inputs(); // term size
    t_sizeOu = useterm.getOMSize(); // term size
 
    int counter = 0; // counter of terms
    int i_max = 0; // max value
    bool sameOu;
        
    int * array_ints; // array of dc values
    array_ints = (int*)malloc(90000*sizeof(int));
    
      
    /* counter intersects */  
    for( t1 = pla->getData()->begin(); t1 != pla->getData()->end(); t1++ ) {
     
       array_ints[counter] = 0;
       useterm = *t1;
    
       for( t2 = pla->getData()->begin(); t2 != pla->getData()->end(); t2++ ) {
                
          /* counts ints for invidual term*/ 
          if(t1 != t2) {
         
             sameOu = true; // default 
             newterm = *t2;
             
             for (int j = 0; j < t_sizeOu; j++ ) {
                /* if terms have at least same one "1" on output */
               	if (newterm.getOM(j) != useterm.getOM(j)) {                      
                    sameOu = false;  
                }
             } 
   
            /* check the neighbors*/         
            if (sameOu == true) {
               int count_another = 0;      
               bool merge = true;
                 
               for (int j = 0; j < t_sizeIn; j++ ) {
                  /* if terms have at least same one "1" on output */
                   if (((newterm.isLit_1(j)) && (useterm.isLit_0(j))) || 
                       ((newterm.isLit_0(j)) && (useterm.isLit_1(j)))) {                       
                     count_another++;     
                     
                  }
                  /* if can't be merge */
                  if (((newterm.isLit_dc(j)) && (!useterm.isLit_dc(j))) || 
                      ((!newterm.isLit_dc(j)) && (useterm.isLit_dc(j)))) {                                            
                     merge = false;     
                  }
               } 
               if ((count_another == 1) && (merge == true)) {
                                   
                  array_ints[counter]++;
                
                  if(i_max < array_ints[counter]) 
                     i_max = array_ints[counter];
               }     
            }                                     
          } 
       }
       counter++;
    }           
  
    int pointer_counter = 0; // pointer counter position
   
    /* create new sort pla */
    for ( int j = 0; j <= i_max; j++ ) {   
         counter = 0;
         for( t2 = pla->getData()->begin(); t2 != pla->getData()->end(); t2++ ) {
               useterm = *t2;          
                   
               /* compare ints with grade term */      
               if (j == array_ints[counter]) {
                    /* save term */
                    res.push_back(useterm);
                    
                    /* change term order position */
                    pointer_counter++;
               }     
               counter++;
         }
    }                
     
   list_result->AppendTermList(res); // save result to termlist       	     
   delete pla;
   pla = list_result; // sort result

   free(array_ints); // deallocate memory (array of ints values)     
}

void Ortho::SortByMergeDec(void)
/*
Term MergeNeighbors Sort decrementally
*/
{        
    PLA use;
    PLA *list_result; // result PLA
    
    list_result = new PLA(pla);	
    
    list<Term> res; // result of operations    
	TermList::const_iterator t1, t2, t3;

	Term newterm;    
   	Term useterm; // working term 
 
    int t_sizeIn = 0; // term output size
    int t_sizeOu = 0; // term output size
 
      
    /* find term input and output size */
    t1 = pla->getData()->begin();
    useterm = *t1;
    t_sizeIn = useterm.inputs(); // term size
    t_sizeOu = useterm.getOMSize(); // term size
 
    int counter = 0; // counter of terms
    int i_max = 0; // max value
    bool sameOu;
        
    int * array_ints; // array of dc values
    array_ints = (int*)malloc(90000*sizeof(int));


    /* counter intersects */  
    for( t1 = pla->getData()->begin(); t1 != pla->getData()->end(); t1++ ) {
     
       array_ints[counter] = 0;
       useterm = *t1;
    
       for( t2 = pla->getData()->begin(); t2 != pla->getData()->end(); t2++ ) {

          /* counts ints for invidual term*/ 
          if(t1 != t2) {
         
             sameOu = true; // default 
             newterm = *t2;
             
             for (int j = 0; j < t_sizeOu; j++ ) {
                /* if terms have at least same one "1" on output */
               	if (newterm.getOM(j) != useterm.getOM(j)) {                      
                    sameOu = false;  
                }
             } 
   
            /* check the neighbors*/         
            if (sameOu == true) {
               int count_another = 0;      
               bool merge = true;
                 
               for (int j = 0; j < t_sizeIn; j++ ) {
                  /* if terms have at least same one "1" on output */
                   if (((newterm.isLit_1(j)) && (useterm.isLit_0(j))) || 
                       ((newterm.isLit_0(j)) && (useterm.isLit_1(j)))) {                       
                     count_another++;     
                     
                  }
                  /* if can't be merge */
                  if (((newterm.isLit_dc(j)) && (!useterm.isLit_dc(j))) || 
                      ((!newterm.isLit_dc(j)) && (useterm.isLit_dc(j)))) {                                            
                     merge = false;     
                  }
               } 
               if ((count_another == 1) && (merge == true)) {
                                   
                  array_ints[counter]++;
                
                  if(i_max < array_ints[counter]) 
                     i_max = array_ints[counter];
               }     
            }                                     
          } 
       }
       counter++;
    }           
  
    
    int pointer_counter = 0; // pointer counter position
    
    /* create new sort pla */
    for ( int j = i_max; 0 <= j; j-- ) {
      
         counter = 0;
         for( t2 = pla->getData()->begin(); t2 != pla->getData()->end(); t2++ ) {
               useterm = *t2;          
                   
               /* compare ints with grade term */      
               if (j == array_ints[counter]) {
                    /* save term */
                    res.push_back(useterm);
                    
                    /* change term order position */
                    pointer_counter++;
               }     
               counter++;
         }
    }                
     
   list_result->AppendTermList(res); // save result to termlist       	     
   delete pla;
   pla = list_result; // sort result

   free(array_ints); // deallocate memory (array of ints values)    
}
		
list<Term> Ortho::TermDisjointSharp( const Term& t1, const Term& t2)
/*  
Term Disjoint sharp
*/
{
	unsigned int j, i;
	list<Term> res; // the result
	Term newterm;
	string om;
	

    /* term t1 intersects term t2? */
    if (!t1.intersects(t2)) {
       /* no */ 
       res.push_back(t2);  
  
       return res;                       
    }
     
    bool copies = false;
  
    /* output for intersection term */ 
    om.assign( t2.getOMSize(), '~' ); 
   	for ( j = 0; j <= t1.getOMSize(); j++ ) {
   
      	if ((t1.isOM_1(j)) && (t2.isOM_1(j))) {
          om[j] = '1';                 
          copies = true; 
        }
    } 
 
    
    /* if terms dont itersect ->  t1 disjoint sharp t2 */   
    for ( j = 0; j <= t2.inputs(); j++ ) {

	if ((t2.isLit_dc(j)) && (!t1.isLit_dc(j))) {
           newterm = t2;
         
           for ( i = 0; i < j; i++ ) {
              if ((newterm.isLit_dc(i)))                   
                 newterm.setLit(i, t1.getLit(i));
           }
           
           if (t1.isLit_0(j)) newterm.setLit_1(j);
           if (t1.isLit_1(j)) newterm.setLit_0(j);
                 
           res.push_back(newterm); 
        }
    }	


    /* create intersection term */
    if(copies == true) {
         newterm = t2;
         newterm.setOM(om); // set output
               
         for ( j = 0; j <= t2.inputs(); j++ ) {
             // set input
             if (t1.isLit_1(j)) newterm.setLit_1(j);
             else if (t1.isLit_0(j)) newterm.setLit_0(j);
         }      
 
         res.push_back(newterm);   
    }
 
    return res;  
}



void Ortho::BackMerge( PLA &pla )
/*
MergeNeighbors with orthogonal terms
*/
{ 
    
    PLA *use_pla; 
    use_pla = new PLA(&pla);
 
    PLA use; 
    PLA *use_res; 
    use_res = new PLA(&pla);	       
	TermList::const_iterator t1, t2, t3, t4;
 
    bool continue_process = true;
      
    while (continue_process) {
 
          t1 = pla.getData()->end();
          t1--;
 
          continue_process = false; // default

          for( t2 = pla.getData()->begin(); t2 != pla.getData()->end(); t2++ ) {

              t2++;
              
              if(t2 != pla.getData()->end()){
                    
                 t2--;
                 use_res = new PLA(&pla);
     
                 use_res->AppendTerm(*t1);
                 use_res->AppendTerm(*t2);
	
	             /* Merge Neighbors terms */
                 use_res->MergeNeighbors();      
                  
                 use = *use_res;          
                 t3 = use.getData()->begin();
                 t3++;
                
                                 
                 /* if term is merged */
                 if(t3 != use.getData()->end()) { 
                     /* no */ 
                     use_pla->AppendTerm(*t2);

                 } else {
                      /* yes */  
                      t3--;    
                      t1 = t3;
                      continue_process = true;  
                 }
                        
               } else {
               t2--;
               }        
            }
            use_pla->AppendTerm(*t1);
            
            pla = *use_pla;         
            use_pla = new PLA(&pla);
     }
}

void Ortho::EliminateUndefined(void)
/*
Remove terms where aren't any "1" in output.
*/
{ 
    PLA *result; // result PLA 
    result = new PLA(pla);            
        
	TermList::const_iterator t1;
    
   	Term useterm; // working term 
    int t_size = 0; // term output size
    bool remove_term;    
      
    /* find term size */
    t1 = pla->getData()->begin();
    useterm = *t1;
    t_size = useterm.getOMSize(); // term size
    
    /* cycle of terms */
    for( t1 = pla->getData()->begin(); t1 != pla->getData()->end(); t1++ ) {
          useterm = *t1;
          
          remove_term = true;
          
          /* does output have "1"? */ 
          for ( int j = 0; j <= t_size; j++ ) {
              if (useterm.isOM_1(j)) {
                remove_term = false;                
              }   
          }
       
          /* if output has "1" */
          if(remove_term == false) {
              result->AppendTerm(*t1);
          }
     }           
   
   delete pla;
   pla = result;
}

void Ortho::Ortho_Hoger( unsigned int op )
/* 
Orthogonalization 
*/
{
    PLA *res;
    PLA pla_end;
    
    PLA use;
    PLA *use_res;
  
    list<Term> result; // the result
    list<Term> result_djs; // the result - after disjoint sharp
    list<Term> r_empty; // empty
    
	TermList::const_iterator t1, t2, t3, t4;
 
    bool continue_process  = true;
    bool end_process  = false;
    
	res = new PLA(&*pla);
	ok_terms = new PLA(&*pla);	
	if (( res == NULL) && (ok_terms == NULL)) 
        throw RuntimeException( "PLA_operations: Not enough memory" );


   /* Eliminate terms where aren't any "1" in output. */
    if (op >= 10) EliminateUndefined();
   
   /* main loop of orthogonalization */   
   while (continue_process) {
    
    /* Term Sort */
    if( op == 1) { SizeSortInc(); }
    if( op == 2) { SizeSortDec(); }
    if( op == 3) { IntersectionSortInc(); }
    if( op == 4) { IntersectionSortDec(); }
    if( op == 5) { IntersectionSortInc(); SizeSortDec(); }
    if( op == 6) { IntersectionSortDec(); SizeSortDec(); }
    if( op == 7) { SortByMergeInc(); }
    if( op == 8) { SortByMergeDec(); }

    if( op > 8) { SizeSortDec(); }
        
       t1 = pla->getData()->begin();
       res = new PLA(&*pla); 
                  
       t3 = pla->getData()->begin(); // auxiliary valuable
       t3++; 
       
       /* orthogonalization remaining terms */
       for( t2 = t3; t2 != pla->getData()->end(); t2++ ) {
           
            use_res = new PLA(&*pla);
            use_res->AppendTerm(*t1);
            use_res->AppendTerm(*t2);
      	
	        use_res->MergeNeighbors(); // Merge Neighbors terms              
       
            use = *use_res;          
            t4 = use.getData()->begin();
            t4++;
               
            end_process = false; // default
   
            /* Did term builder collect terms? If not = run term disjoint sharp */
            if(t4 != use.getData()->end()) {     
                 result_djs = r_empty; // assign empty value   
                 result_djs = TermDisjointSharp(*t1,*t2); // disjoint sharp for term t1 and t2
                 res->AppendTermList(result_djs);
                 end_process = true;            
           } else {
              /* after MergeNeighbors and RemoveAbsorbed */
              t4--;
              Term ts1, ts2;
              ts1 = **&t1;
              ts2 = **&t4;

              /* if size of term isn't same */
              if(ts1.getDim() != ts2.getDim()) {    

                   t1 = t4;
                   t2 =  pla->getData()->begin();
  
                   res = new PLA(&*pla); 
                  
              } 
           }  
       } 
             
        ok_terms->AppendTerm(*t1);
        
        /* MergeNeighbors for complete terms*/
        if( op >= 9 ) {
             BackMerge(*ok_terms);
        }
        
        pla = res; // replace old PLA
   
        t3 = pla->getData()->begin(); // auxiliary valuable
        t3++; 

        /* Is next line end? */
        if(t3 == pla->getData()->end()) {
       
              /* if last operations was disjoint sharp */
              if(end_process == true) {
                 t3--; 
                 
                 /* is result exists? */
                 if(t3 != pla->getData()->end()) {
                     ok_terms->AppendTerm(*t3);            
                 }
              } 
              //ok_terms->AppendTermList(result); // save result of orthogonalization   
              continue_process = false; // set end of iterations
        }
   }     
}

// ==================== End Hoger ====================

bool Ortho::IsOrthogonal( PLA *_pla )
/*
    Checks if the PLA is orthogonal
*/
{
    list<Term>::const_iterator ti1, ti2;

    for ( ti1 = _pla->getData()->begin(); ti1 != _pla->getData()->end(); ti1++ )
        for ( ti2 = ti1; ti2 != _pla->getData()->end(); ti2++ )
            if ( ti2 != ti1 )
                if ( (*ti1).intersects(*ti2) )
					return false;
    return true;
}

PLA *Ortho::Run( const unsigned int method )
{
	switch (method) {
		case 1: Orthogonalize();
			break;
		case 2: Orthogonalize2();
			break;
		case 3: Orthogonalize3();
			break;
		case 4: Orthogonalize4();
			break;
		case 5: Orthogonalize5();
			break;
		case 6: Ortho_Hoger(10);
			break;
		default: throw RuntimeException( "Ortho:: Unknown orthogonalization method" );
	}

	ok_terms->setNonUnique();
	ok_terms->MergeNeighbors();
	ok_terms->MergeTerms();

	if ( !IsOrthogonal( ok_terms ) ) throw RuntimeException( "Ortho::Something's wrong - resulting PLA is not orthogonal" );
	return ok_terms;
}

