//---------------------------------------------------------------------------

#include "stdafx.h"

using namespace std;

#ifdef _MSC_VER
#pragma hdrstop
#endif

#include "../Kernel/kernelincludes.h"
#include "../Kernel/utils.h"

#include "Blif.h"

//---------------------------------------------------------------------------


Blif::Blif()
{
}

Blif::~Blif()
{
}

Blif::Blif( const Blif *orig )
{
	inames = orig->inames;
	onames = orig->onames;
	model = orig->model;
}

unsigned int Blif::inputs(void) const
{
	return inames.size();
}

unsigned int Blif::outputs(void) const
{
	return onames.size();
}

Blif::Blif( const vector<string> _inames, const vector<string> _onames )
{
	inames = _inames;
	onames = _onames;
}

void Blif::clear(void)
{
	inames.clear();
	onames.clear();
	data.clear();
}

void Blif::LoadBlif(const char *fn)
{
    unsigned int pos;
    ifstream f;
    string s, s1, gatename, oname;
    char line[MAXLINE];
    Cube t;
    unsigned int i, j, k;
	vector<string> tmpnames;
	vector<string>::iterator tmpnames_it;

    this->clear();

    f.open(fn);
    if (!f.is_open()) throw IOException( "Cannot open file: " + string(fn) );

			// read the header
    GetLine(f, line);
	while ( !f.eof() ) {
        pos = 0;
		if ( line[0] != '#' && line[0] != 0 ) {
			if ( line[0] == '.' ) {
			   s = GetToken(line, pos);
			   if ( s == ".model" ) {
					s = GetToken(line, pos);
					model = s;
			   } else if ( s == ".inputs" ) {
					s = GetToken(line, pos);
					while ( !s.empty() ) {
						inames.push_back(s);
						s = GetToken(line, pos);
					}
			   } else if ( s == ".outputs" ) {
					s = GetToken(line, pos);
					while ( !s.empty() ) {
						onames.push_back(s);
						s = GetToken(line, pos);
					}
			   } else if ( s == ".default_input_arrival" ) {
					s = GetToken(line, pos);
					s = GetToken(line, pos);
			   } else if ( s == ".default_output_required" ) {
					s = GetToken(line, pos);
					s = GetToken(line, pos);
			   } else if ( s == ".default_input_drive" ) {
					s = GetToken(line, pos);
					s = GetToken(line, pos);
			   } else if ( s == ".default_output_load" ) {
					s = GetToken(line, pos);
			   } else if ( s == ".default_max_input_load" ) {
					s = GetToken(line, pos);
			   } else if ( s == ".wire_load_slope" ) {
					s = GetToken(line, pos);

			   } else if ( s == ".names" || s == ".latch" ) break;		// end of the header
			   else if ( s == ".gate" ) break;		// end of the header
			   else if ( s == ".end" ) break;
			   else if ( s == ".exdc" ) break;
			   else throw IOException( "Bad or damaged Blif file: " + string(fn) );
			}
		}
		GetLine(f, line);
	}
			// read nodes
	while ( !f.eof() ) {
        pos = 0;
		if ( line[0] != '#' && line[0] != 0 ) {
			if ( line[0] == '.' ) {
			   s = GetToken(line, pos);
			   if ( s == ".end" ) break;
			   else if ( s == ".exdc" ) break;
			   else if ( s == ".names" ) {
					tmpnames.clear();
					s = GetToken(line, pos);
					while ( !s.empty() ) {
						tmpnames.push_back(s);
						s = GetToken(line, pos);
					}
					if ( tmpnames.size() == 0 ) throw IOException( "Bad or damaged Blif file: " + string(fn) );
					if ( tmpnames.size() == 1 ) {	// constant 0 or 1 node
						data.push_back(SoPLA(1));
					} else data.push_back(SoPLA(tmpnames.size()-1));
					SoPLA& node = data.back();
					node.setName(tmpnames.back());		// set node names
					tmpnames.pop_back();	// remove the last name (output)
					i = 0;
					for ( tmpnames_it = tmpnames.begin(); tmpnames_it != tmpnames.end(); tmpnames_it++ )
						node.setIName(i++, (*tmpnames_it));	

						// parse the node PLA
					do {		
						GetLine(f, line);
						if ( line[0] != 0 && line[0] != '.' ) {
							pos = 0;				
							s = GetToken(line, pos);
							if ( s.length() != node.inputs() && node.inputs() != 0 ) throw IOException( "Bad or damaged Blif file: " + string(fn) );
							if ( tmpnames.size() == 0 ) {						
								t.clear();
								t.setCube("-");
								if ( s == "1" )	node.setType(true);
								else if ( s == "0" ) node.setType(false);
							} else {
								s1.clear();
								for ( i = 0; i < node.inputs(); i++ )
								  if ( s[i] == '0' || s[i] == '1' ) s1 += s[i];
								  else if ( s[i] == '-' || s[i] == '2' ) s1 += '-';
								  else throw IOException( "Bad or damaged Blif file: " + string(fn) );

								t.clear();
								t.setCube(s1);

								s = GetToken(line, pos);
								if ( s.length() != 1 ) throw IOException( "Bad or damaged Blif file: " + string(fn) );
								if ( node.getTerms() == 0 ) {
									if ( s == "1" )	node.setType(true);
									else if ( s == "0" ) node.setType(false);
									else throw IOException( "Bad or damaged Blif file: " + string(fn) );
								} else if ( (s != "0" && s != "1") || (s == "0" && node.getType()) || (s == "1" && !node.getType()) ) throw IOException( "Bad or damaged Blif file: " + string(fn) );
							}
							node.AppendTerm(t);
						} else if ( line[0] == '.' && node.getTerms() == 0 ) {
							t.clear();
							t.setCube("-");
							node.setType(false);
							node.AppendTerm(t);
						}

					} while ( !f.eof() && line[0] != '.' );
					continue;
			   } else if ( s == ".gate" ) {
				    s = GetToken(line, pos);
					j = s[s.length()-1];
					if ( j > 48+9 || j < 48 ) throw IOException( "Bad or damaged Blif file: " + string(fn) + ". Unknown gate" );
					gatename = s.substr( 0, s.length()-1);
					j -= 48;
					tmpnames.clear();
 					for ( i = 0; i < j; i++ ) {
						s = GetToken(line, pos);
						if ( s.empty() ) IOException( "Bad or damaged Blif file: " + string(fn) );
						k = 0;
						while ( s[k] != '=' ) {
							k++;
							if ( k == s.length()-2 ) IOException( "Bad or damaged Blif file: " + string(fn) );
						}
						tmpnames.push_back(s.substr(k+1, s.length()-k-1) );	
					}
					s = GetToken(line, pos);
					if ( s.empty() ) IOException( "Bad or damaged Blif file: " + string(fn) );
					k = 0;
					while ( s[k] != '=' ) {
						k++;
						if ( k == s.length()-2 ) IOException( "Bad or damaged Blif file: " + string(fn) );
					}
					oname = s.substr(k+1, s.length()-k-1);
					data.push_back(SoPLA(j));
					SoPLA& node = data.back();
					node.setName(oname);
					i = 0;
					for ( tmpnames_it = tmpnames.begin(); tmpnames_it != tmpnames.end(); tmpnames_it++ )
						node.setIName(i++, (*tmpnames_it));	
					node.setGateName( gatename );
			   } else if ( s == ".latch" ) {
					data.push_back(SoPLA(1));
					SoPLA& node = data.back();
					s = GetToken(line, pos);
					if ( s == "" ) throw IOException( "Bad or damaged Blif file: " + string(fn) );
					node.setIName( 0, s );
					s = GetToken(line, pos);
					if ( s == "" ) throw IOException( "Bad or damaged Blif file: " + string(fn) );
					node.setName(s);
					node.setDFF(true);
			   } else throw IOException( "Bad or damaged Blif file: " + string(fn) );
			}
		}
		GetLine(f, line);
    }
	f.close();
}

void Blif::LoadBench( const char *fn )
// loads Blif from Bench file
{
    ifstream f;
    string s, s1, signame, gatename, oname;
    char line[MAXLINE];
    unsigned int i, j;
	Cube t;
	list<SoPLA>::iterator node;
	vector<string> nd_inames;

		
	this->clear();

    f.open(fn);
    if (!f.is_open()) throw IOException( "Cannot open file: " + string(fn) );

	while ( !f.eof() ) {
		GetLine(f, line);
		if ( line[0] == 0 ) continue;
		if ( line[0] == '#' ) continue;
		s = line;
		if ( strnicmp(line, "INPUT(", 6) == 0 ) {
			s1 = s.substr(6, s.find(")", 0)-6);
			inames.push_back(RemoveSpaces(s1));
		}
		else if ( strnicmp(line, "OUTPUT(", 7) == 0 ) {
			s1 = s.substr(7, s.find(")", 0)-7);
			onames.push_back(RemoveSpaces(s1));
		}
		else if ( s.find( "=", 1 ) != std::string::npos ) {
			i = s.find( "=", 1 );
			signame = RemoveSpaces( s.substr(0, i-1) );
			j = s.find("(", i+1);
			gatename = RemoveSpaces( s.substr(i+1, j-i-1));
			to_upper(gatename);
			i = j+1;
			nd_inames.clear();
			while ( s[i] != ')' ) {
				if ( i == s.length() ) throw IOException( "Cannot open file: " + string(fn) );
				j = s.find(",", i );
				if ( j == std::string::npos ) j = s.find(")", i );
				if ( j == std::string::npos ) throw IOException( "Cannot open file: " + string(fn) );
				nd_inames.push_back(RemoveSpaces(s.substr(i, j-i)));
				i = j+1;
				if ( i == s.length() ) break;
			}
			data.push_back(SoPLA(nd_inames.size()));
			node = data.end();
			node --;
			node->setName(signame);
			node->setINames(nd_inames);
			node->RecognizeFunction(gatename, nd_inames.size() );
		}
	}
}

void Blif::SaveBlif( const char *fn )
// Saves the network to a BLIF file
{
	ofstream f;
	unsigned int i;
	vector<string>::const_iterator name_it;
	list<SoPLA>::iterator node_it;
	CubeList::iterator ti;

    f.open(fn);
	if (!f.is_open()) throw IOException( "Blif::SaveBlif: Cannot open file: " + string(fn) );

		// write header
	if ( !model.empty() ) f << ".model " << model << endl;
	f << ".inputs";
	for ( name_it = inames.begin(); name_it != inames.end(); name_it++ )
		f << " " << *name_it;
	f << endl;
	f << ".outputs";
	for ( name_it = onames.begin(); name_it != onames.end(); name_it++ )
		f << " " << *name_it;
	f << endl;

		// write nodes
	for ( node_it = data.begin(); node_it != data.end(); ++node_it ) {
		if ( node_it->inputs() == 0 ) continue;		// empty node. It may happen sometimes.
		f << ".names";
		if ( node_it->inputs() == 1 && node_it->getTerms() == 1 && node_it->getConstDataRef().front().getCube() == "-" ) {   // constant node
			f << " " << node_it->getName() << endl;
			f << " " << ((node_it->getType())?"1":"0") << endl;
		} else {
			for ( i = 0; i < node_it->inputs(); i++ )
				if ( !node_it->getIName(i).empty() || node_it->getData()->front().getCube() == "-" ) f << " " << node_it->getIName(i);
				else throw RuntimeException( "Blif::SaveBlif: Empty node input name" );
			if ( node_it->getName().empty() ) throw RuntimeException( "Blif::SaveBlif: Empty node output name" );
			f << " " << (*node_it).getName() << endl;
			if ( node_it->getData()->front().getCube() != "-" ) {
				for ( ti = node_it->getData()->begin(); ti != (*node_it).getData()->end(); ++ti )
					f << ti->getCube() << " " << ((node_it->getType())?"1":"0") << endl;
			} else f << " " << ((node_it->getType())?"1":"0") << endl;
		}
	}
	f << ".end" << endl;
}

void Blif::SaveBench( const char *fn )
// Saves the network to a BLIF file
{
	ofstream f;
	unsigned int i;
	vector<string>::const_iterator name_it;
	list<SoPLA>::iterator node_it;
	CubeList::iterator ti;

    f.open(fn);
	if (!f.is_open()) throw IOException( "Blif::SaveBench: Cannot open file: " + string(fn) );

		// write header
	for ( name_it = inames.begin(); name_it != inames.end(); name_it++ )
		f << "INPUT(" << *name_it << ")" << endl;
	f << endl;
	for ( name_it = onames.begin(); name_it != onames.end(); name_it++ )
		f << "OUTPUT(" << *name_it << ")" << endl;
	f << endl;

		// write nodes
	for ( node_it = data.begin(); node_it != data.end(); node_it++ ) {
		if ( (*node_it).getName().empty() ) IOException( "Blif::SaveBench: output name not specified" );
		f << (*node_it).getName() << " = ";
		if ( (*node_it).getGateName().empty() ) IOException( "Blif::SaveBench: gate type not specified" );
		f << (*node_it).getGateName() << "(";
		if ( (*node_it).getIName(0).empty() ) IOException( "Blif::SaveBench: input name not specified" );
		f << (*node_it).getIName(0);
		for ( i = 1; i < (*node_it).inputs(); i++ )
			if ( (*node_it).getIName(i).empty() ) IOException( "Blif::SaveBench: input name not specified" );
			else f << ", " << (*node_it).getIName(i);
		f << ")" << endl;
	}
}

void Blif::AppendPLA( const SoPLA *_pla )
{
	data.push_back( *_pla );
}

void Blif::Splice( list<SoPLA> &plas )
{
	data.splice(data.end(), plas);
}

void Blif::Splice( Blif &blif )
{
	data.splice(data.end(), blif.data);
}

void Blif::AppendBlif( const Blif *_blif )
// Joins data from two Blifs. Input and output names are not considered!
{
	list<SoPLA>::const_iterator pla_it;

	for ( pla_it = _blif->data.begin(); pla_it != _blif->data.end(); pla_it++ )
		this->AppendPLA(&*pla_it);
}

unsigned int Blif::getLiterals(void)
{
	unsigned int cnt = 0;
	list<SoPLA>::iterator pla_it;

	for ( pla_it = data.begin(); pla_it != data.end(); ++pla_it )
		cnt += (*pla_it).getLiterals();
	return cnt;
}

unsigned int Blif::getTerms(void) const
{
	unsigned int cnt = 0;
	list<SoPLA>::const_iterator pla_it;

	for ( pla_it = data.begin(); pla_it != data.end(); ++pla_it )
		cnt += pla_it->getTerms();
	return cnt;
}

unsigned int Blif::getGEs(void)
{
	unsigned int cnt = 0;
	list<SoPLA>::iterator pla_it;

	for ( pla_it = data.begin(); pla_it != data.end(); pla_it++ )
		cnt += (*pla_it).getGEs();
	return cnt;
}

unsigned int Blif::GetLevels(void)
	// returns the number of BLIF levels (longest path)
{
	list<SoPLA>::iterator pla_it, pla_it2;
	vector<string>::iterator s_it;
	bool change;
	unsigned int my_max = 1;
	unsigned int i;
	
	for ( pla_it = data.begin(); pla_it != data.end(); pla_it++ )
		(*pla_it).tmp = 0;
	for ( s_it = inames.begin(); s_it != inames.end(); s_it++ )
		for ( pla_it = data.begin(); pla_it != data.end(); pla_it++ )
			for ( i = 0; i < (*pla_it).inputs(); i++ )
				if ( (*s_it) == (*pla_it).getIName(i) ) (*pla_it).tmp = 1;
	do {
		change = false;
		for ( pla_it = data.begin(); pla_it != data.end(); pla_it++ )
			if ( (*pla_it).tmp != 0 )
				for ( pla_it2 = data.begin(); pla_it2 != data.end(); pla_it2++ )
					if ( pla_it != pla_it2 )
//						if ( (*pla_it2).inputs() > 1 )		// just for asynch! Do not count trivial nodes
						for ( i = 0; i < (*pla_it2).inputs(); i++ )
							if ( (*pla_it).getName() == (*pla_it2).getIName(i) ) 
								if ( (*pla_it2).tmp <= (*pla_it).tmp ) {
									(*pla_it2).tmp = (*pla_it).tmp + 1;
									if ( (*pla_it2).tmp > my_max ) my_max = (*pla_it2).tmp;
									change = true;	
									break;
								}
	} while (change);
	return my_max;
}

unsigned int Blif::CountNodeLevels( list<SoPLA>::iterator pla_it, unsigned int k, bool notriv )
{
	unsigned int lev = 0, maxlev = 0;
	list<Cube>::iterator ti;

	if ( notriv && pla_it->isTrivial() ) return 0;
	else if ( k == 0 ) {			// no tech limit
		if ( pla_it->getTerms() == 1 ) return 1;	// only one AND gate
		else if ( pla_it->getTerms() == pla_it->inputs() ) return 1;	// only one OR gate
		else return 2;
	} else {
		// for ANDs
		if ( pla_it->getTerms() != pla_it->inputs() ) {
			for ( ti = pla_it->getData()->begin(); ti != pla_it->getData()->end(); ++ti ) {
				if ( ti->getLiterals() > 1 ) {
					lev = 2*((unsigned int)ceil((float)(ti->getLiterals() - 1)/float(k-1)))-1;
					if ( lev > maxlev ) maxlev = lev;
				}
			}
		} else lev = 0;
		// for OR
		if ( pla_it->getTerms() > 1 )
			lev = 2*((unsigned int)ceil((float)(pla_it->getTerms() - 1)/float(k-1)))-1;
		else lev = 0;
		return maxlev + lev;
	}
}

unsigned int Blif::GetNAND_NANDLevels(unsigned int k, bool notriv)
	// returns the number of levels of the AND-OR implementation
	// k is the technology limit - AND and OR gates are decomposed
	// k = 0 => no techlology limit
	// notriv = true => do not count trivial nodes (buffers, inverters)
{
	list<SoPLA>::iterator pla_it, pla_it2;
	vector<string>::iterator s_it;
	bool change;
	unsigned int lev, my_max = 0;
	unsigned int i;
	
		// set levels of gates fed by PIs
	for ( pla_it = data.begin(); pla_it != data.end(); ++pla_it )
		pla_it->tmp = UINT_MAX;
	
	for ( pla_it = data.begin(); pla_it != data.end(); ++pla_it )
		for ( s_it = inames.begin(); s_it != inames.end(); ++s_it )
			for ( i = 0; i < pla_it->inputs(); i++ )
				if ( (*s_it) == pla_it->getIName(i) ) {
					lev = CountNodeLevels(pla_it, k, notriv);
					pla_it->tmp = lev;
					if ( my_max < lev ) my_max = lev;
				}
	do {
		change = false;
		for ( pla_it = data.begin(); pla_it != data.end(); ++pla_it )
			if ( pla_it->tmp != UINT_MAX )
				for ( pla_it2 = data.begin(); pla_it2 != data.end(); ++pla_it2 )
					if ( pla_it != pla_it2 )
						for ( i = 0; i < pla_it2->inputs(); i++ )
							if ( pla_it->getName() == pla_it2->getIName(i) ) {
								lev = CountNodeLevels(pla_it2, k, notriv);
								if ( pla_it2->tmp == UINT_MAX || pla_it2->tmp < pla_it->tmp + lev ) {
									pla_it2->tmp = pla_it->tmp + lev;
									if ( pla_it2->tmp > my_max ) my_max = pla_it2->tmp;
									change = true;
									break;
								}
							}
	} while (change);
	return my_max;
}

unsigned int Blif::GetNonTrivNodes(void)
{
	unsigned int nd = 0;
	list<SoPLA>::iterator pla_it;

	for ( pla_it = data.begin(); pla_it != data.end(); ++pla_it )
		if ( !pla_it->isTrivial() ) nd++;
	return nd;
}

string Blif::getIName( const unsigned int i ) const
{
	if ( i >= inames.size() ) throw RuntimeException("Bilf:: getIName: wrong name index");
	return inames[i];
}

string Blif::getOName( const unsigned int i ) const
{
	if ( i >= onames.size() ) throw RuntimeException("Bilf:: getOName: wrong name index");
	return onames[i];
}
