//---------------------------------------------------------------------------

#ifndef BlifH
#define BlifH
                       
class Blif {
  private:
	string model;					// name of the model
	vector<string> inames;			// names of input ports. Must be defined
	vector<string> onames;			// names of output ports. Must be defined
	list<SoPLA> data;				    // here are the PLAs

	void clear(void);				// clears all the structure

  public:
	unsigned int CountNodeLevels( list<SoPLA>::iterator pla_it, unsigned int k, bool notriv );		// for GetNAND_NANDLevels

	Blif(void);
	Blif( const Blif *orig );			// non-copying constructor
	Blif( const vector<string> _inames, const vector<string> _onames );
    ~Blif(void);

	void LoadBlif( const char *fn );	// loads Blif from file
	void LoadBench( const char *fn );	// loads Blif from Bench file
	void SaveBlif( const char *fn );	// save as Blif
	void SaveBench( const char *fn );	// save as Bench

		// getters & setters
	unsigned int inputs(void) const;
	unsigned int outputs(void) const;
	list<SoPLA> *getData(void) { return &data; };
	const list<SoPLA> &getConstDataRef(void) const { return data; };
	string GetModel(void) const { return model; };
	void SetModel(const string _model) { model = _model; };
	unsigned int GetNodes(void) const { return data.size(); };
	vector<string> getINames(void) const { return inames;};
	vector<string> getONames(void) const { return onames;};
	string getIName( const unsigned int i ) const;
	string getOName( const unsigned int i ) const;

		// info
	unsigned int getLiterals(void);
	unsigned int getTerms(void) const;
	unsigned int getGEs(void);
	unsigned int GetLevels(void);		// returns the number of BLIF levels (longest path)
	unsigned int GetNAND_NANDLevels(bool notriv) { return(GetNAND_NANDLevels(0, notriv)); };		// returns the number of levels of the NAND-NAND implementation
	unsigned int GetNAND_NANDLevels(unsigned int k, bool notriv);		// returns the number of levels of the NAND-NAND implementation for technology dependent synthesis
	unsigned int GetNonTrivNodes(void);	// returns the number of non-trivial (not inv, buff, const) nodes

		// data manipulation
	void AppendPLA( const SoPLA *_pla );
	void Splice( list<SoPLA> &plas );
	void Splice( SoPLA *_pla );
	void Splice( Blif &blif );
	void AppendBlif( const Blif *_blif );	// Joins data from two Blifs. Input and output names are not considered!
};

//---------------------------------------------------------------------------
#endif
